""" Generic operations on symbolic expresions and data structures """
import sympy as sp

def _float(fpprec, f):
    """ function wrapper for conditional casting to python float """
    if fpprec == 'fixed':
        def newfunc(*args, **kwargs):
            return float(f(*args, **kwargs))
    else:
        newfunc = f
    return newfunc

def spfloat(f, prec):
    """ creates a sympy float with specified precision from a python float """
    from decimal import Decimal
    if isinstance(f, float):
        x, y = Decimal(repr(f)).as_integer_ratio()
        return sp.N(x, prec)/sp.N(y, prec)
    else:
        return f

def flatten(x):
    """ Return a flat list from a deeply nested iterables """
    from collections import Iterable
    return (a for i in x for a in flatten(i)) if isinstance(x, Iterable) else [x]

def uniq1(seq):
    """ filter out duplicated elements and return the unique ones, faster """
    seen = set()
    seen_add = seen.add
    return [x for x in seq if not (x in seen or seen_add(x))]

def uniq2(seq):
    """ filter out duplicated elements and return the unique ones, slower """
    res = []
    for test in seq:
        if not any(test == c for c in res):
            res.append(test)
    return res

def get_partial_derivs(function, variables, n, flatlist=True):
    """ Evaluate the first n partial derivatives of a function with respect to
        specified variables and return a list of unique derivatives """
    
    ders = [uniq1(sp.derive_by_array(function, variables))]
    for j in range(1, n):
        ders.append(uniq1(sp.derive_by_array(ders[j-1], variables)))
    return flatten(ders) if flatlist else ders

def get_all_partial_derivs(function, variables, n, flatlist=True):
    """ Evaluate the first n partial derivatives of a function with respect to
        specified variables and return a flat iterable or a list of tensors """
    ders = [sp.derive_by_array(function, variables)]
    for j in range(1, n):
        ders.append(sp.derive_by_array(ders[j-1], variables))
    return flatten(ders) if flatlist else ders

def lambdify_long(symbs, expr, *args, **kwargs):
    """ symbs: an iterable of symbols, expr: expression with these symbols """
    from types import GeneratorType
    from collections import Iterable
    from itertools import chain
    dvec = sp.DeferredVector('x')
    repl = {q: dvec[i] for i, q in enumerate(symbs)}
    if isinstance(expr, (Iterable, GeneratorType, chain)):
        sexpr = [e.xreplace(repl) for e in expr]
    else:
        sexpr = expr.xreplace(repl)
    return sp.lambdify(dvec, sexpr, *args, **kwargs)

def validate_numeric(fpprec, numeric):
    """ validate consistency between fpprec and numeric parameters """
    if fpprec != 'fixed':
        assert numeric['tool'] in ['subs', 'lambdify']
        if numeric['tool'] == 'lambdify':
            mods = numeric['modules']
            if isinstance(mods, list):
                assert all(f in ['mpmath', 'sympy'] for f in mods)
            else:
                assert mods in ['mpmath', 'sympy']
